@@ -1,18 +1,24 @@
 """Unit tests for the copy module."""
 
 import copy
-import copy_reg
+import copyreg
 import weakref
+import abc
+from operator import le, lt, ge, gt, eq, ne
 
 import unittest
-from test import test_support
+from test import support
+
+order_comparisons = le, lt, ge, gt
+equality_comparisons = eq, ne
+comparisons = order_comparisons + equality_comparisons
 
 class TestCopy(unittest.TestCase):
 
     # Attempt full line coverage of copy.py from top to bottom
 
     def test_exceptions(self):
-        self.assertTrue(copy.Error is copy.error)
+        self.assertIs(copy.Error, copy.error)
         self.assertTrue(issubclass(copy.Error, Exception))
 
     # The copy() method
@@ -43,32 +49,38 @@ class TestCopy(unittest.TestCase):
             return (C, (obj.foo,))
         x = C(42)
         self.assertRaises(TypeError, copy.copy, x)
-        copy_reg.pickle(C, pickle_C, C)
+        copyreg.pickle(C, pickle_C, C)
         y = copy.copy(x)
 
     def test_copy_reduce_ex(self):
         class C(object):
             def __reduce_ex__(self, proto):
+                c.append(1)
                 return ""
             def __reduce__(self):
-                raise test_support.TestFailed, "shouldn't call this"
+                self.fail("shouldn't call this")
+        c = []
         x = C()
         y = copy.copy(x)
-        self.assertTrue(y is x)
+        self.assertIs(y, x)
+        self.assertEqual(c, [1])
 
     def test_copy_reduce(self):
         class C(object):
             def __reduce__(self):
+                c.append(1)
                 return ""
+        c = []
         x = C()
         y = copy.copy(x)
-        self.assertTrue(y is x)
+        self.assertIs(y, x)
+        self.assertEqual(c, [1])
 
     def test_copy_cant(self):
         class C(object):
             def __getattribute__(self, name):
                 if name.startswith("__reduce"):
-                    raise AttributeError, name
+                    raise AttributeError(name)
                 return object.__getattribute__(self, name)
         x = C()
         self.assertRaises(copy.Error, copy.copy, x)
@@ -82,11 +94,14 @@ class TestCopy(unittest.TestCase):
             pass
         def f():
             pass
-        tests = [None, 42, 2L**100, 3.14, True, False, 1j,
-                 "hello", u"hello\u1234", f.func_code,
-                 NewStyle, xrange(10), Classic, max]
+        class WithMetaclass(metaclass=abc.ABCMeta):
+            pass
+        tests = [None, 42, 2**100, 3.14, True, False, 1j,
+                 "hello", "hello\u1234", f.__code__,
+                 b"world", bytes(range(256)),
+                 NewStyle, range(10), Classic, max, WithMetaclass]
         for x in tests:
-            self.assertTrue(copy.copy(x) is x, repr(x))
+            self.assertIs(copy.copy(x), x)
 
     def test_copy_list(self):
         x = [1, 2, 3]
@@ -104,8 +119,8 @@ class TestCopy(unittest.TestCase):
         class C:
             def __init__(self, foo):
                 self.foo = foo
-            def __cmp__(self, other):
-                return cmp(self.foo, other.foo)
+            def __eq__(self, other):
+                return self.foo == other.foo
         x = C(42)
         self.assertEqual(copy.copy(x), x)
 
@@ -115,8 +130,8 @@ class TestCopy(unittest.TestCase):
                 self.foo = foo
             def __copy__(self):
                 return C(self.foo)
-            def __cmp__(self, other):
-                return cmp(self.foo, other.foo)
+            def __eq__(self, other):
+                return self.foo == other.foo
         x = C(42)
         self.assertEqual(copy.copy(x), x)
 
@@ -126,8 +141,8 @@ class TestCopy(unittest.TestCase):
                 self.foo = foo
             def __getinitargs__(self):
                 return (self.foo,)
-            def __cmp__(self, other):
-                return cmp(self.foo, other.foo)
+            def __eq__(self, other):
+                return self.foo == other.foo
         x = C(42)
         self.assertEqual(copy.copy(x), x)
 
@@ -137,8 +152,8 @@ class TestCopy(unittest.TestCase):
                 self.foo = foo
             def __getstate__(self):
                 return {"foo": self.foo}
-            def __cmp__(self, other):
-                return cmp(self.foo, other.foo)
+            def __eq__(self, other):
+                return self.foo == other.foo
         x = C(42)
         self.assertEqual(copy.copy(x), x)
 
@@ -148,8 +163,8 @@ class TestCopy(unittest.TestCase):
                 self.foo = foo
             def __setstate__(self, state):
                 self.foo = state["foo"]
-            def __cmp__(self, other):
-                return cmp(self.foo, other.foo)
+            def __eq__(self, other):
+                return self.foo == other.foo
         x = C(42)
         self.assertEqual(copy.copy(x), x)
 
@@ -161,10 +176,13 @@ class TestCopy(unittest.TestCase):
                 return self.foo
             def __setstate__(self, state):
                 self.foo = state
-            def __cmp__(self, other):
-                return cmp(self.foo, other.foo)
+            def __eq__(self, other):
+                return self.foo == other.foo
         x = C(42)
         self.assertEqual(copy.copy(x), x)
+        # State with boolean value is false (issue #25718)
+        x = C(0.0)
+        self.assertEqual(copy.copy(x), x)
 
     # The deepcopy() method
 
@@ -180,9 +198,9 @@ class TestCopy(unittest.TestCase):
         x = [x, x]
         y = copy.deepcopy(x)
         self.assertEqual(y, x)
-        self.assertTrue(y is not x)
-        self.assertTrue(y[0] is not x[0])
-        self.assertTrue(y[0] is y[1])
+        self.assertIsNot(y, x)
+        self.assertIsNot(y[0], x[0])
+        self.assertIs(y[0], y[1])
 
     def test_deepcopy_issubclass(self):
         # XXX Note: there's no way to test the TypeError coming out of
@@ -191,8 +209,8 @@ class TestCopy(unittest.TestCase):
         # type.
         class Meta(type):
             pass
-        class C:
-            __metaclass__ = Meta
+        class C(metaclass=Meta):
+            pass
         self.assertEqual(copy.deepcopy(C), C)
 
     def test_deepcopy_deepcopy(self):
@@ -216,32 +234,38 @@ class TestCopy(unittest.TestCase):
             return (C, (obj.foo,))
         x = C(42)
         self.assertRaises(TypeError, copy.deepcopy, x)
-        copy_reg.pickle(C, pickle_C, C)
+        copyreg.pickle(C, pickle_C, C)
         y = copy.deepcopy(x)
 
     def test_deepcopy_reduce_ex(self):
         class C(object):
             def __reduce_ex__(self, proto):
+                c.append(1)
                 return ""
             def __reduce__(self):
-                raise test_support.TestFailed, "shouldn't call this"
+                self.fail("shouldn't call this")
+        c = []
         x = C()
         y = copy.deepcopy(x)
-        self.assertTrue(y is x)
+        self.assertIs(y, x)
+        self.assertEqual(c, [1])
 
     def test_deepcopy_reduce(self):
         class C(object):
             def __reduce__(self):
+                c.append(1)
                 return ""
+        c = []
         x = C()
         y = copy.deepcopy(x)
-        self.assertTrue(y is x)
+        self.assertIs(y, x)
+        self.assertEqual(c, [1])
 
     def test_deepcopy_cant(self):
         class C(object):
             def __getattribute__(self, name):
                 if name.startswith("__reduce"):
-                    raise AttributeError, name
+                    raise AttributeError(name)
                 return object.__getattribute__(self, name)
         x = C()
         self.assertRaises(copy.Error, copy.deepcopy, x)
@@ -255,76 +279,106 @@ class TestCopy(unittest.TestCase):
             pass
         def f():
             pass
-        tests = [None, 42, 2L**100, 3.14, True, False, 1j,
-                 "hello", u"hello\u1234", f.func_code,
-                 NewStyle, xrange(10), Classic, max]
+        tests = [None, 42, 2**100, 3.14, True, False, 1j,
+                 "hello", "hello\u1234", f.__code__,
+                 NewStyle, range(10), Classic, max]
         for x in tests:
-            self.assertTrue(copy.deepcopy(x) is x, repr(x))
+            self.assertIs(copy.deepcopy(x), x)
 
     def test_deepcopy_list(self):
         x = [[1, 2], 3]
         y = copy.deepcopy(x)
         self.assertEqual(y, x)
-        self.assertTrue(x is not y)
-        self.assertTrue(x[0] is not y[0])
+        self.assertIsNot(x, y)
+        self.assertIsNot(x[0], y[0])
 
     def test_deepcopy_reflexive_list(self):
         x = []
         x.append(x)
         y = copy.deepcopy(x)
-        self.assertRaises(RuntimeError, cmp, y, x)
-        self.assertTrue(y is not x)
-        self.assertTrue(y[0] is y)
+        for op in comparisons:
+            self.assertRaises(RuntimeError, op, y, x)
+        self.assertIsNot(y, x)
+        self.assertIs(y[0], y)
         self.assertEqual(len(y), 1)
 
+    def test_deepcopy_empty_tuple(self):
+        x = ()
+        y = copy.deepcopy(x)
+        self.assertIs(x, y)
+
     def test_deepcopy_tuple(self):
         x = ([1, 2], 3)
         y = copy.deepcopy(x)
         self.assertEqual(y, x)
-        self.assertTrue(x is not y)
-        self.assertTrue(x[0] is not y[0])
+        self.assertIsNot(x, y)
+        self.assertIsNot(x[0], y[0])
+
+    def test_deepcopy_tuple_of_immutables(self):
+        x = ((1, 2), 3)
+        y = copy.deepcopy(x)
+        self.assertIs(x, y)
 
     def test_deepcopy_reflexive_tuple(self):
         x = ([],)
         x[0].append(x)
         y = copy.deepcopy(x)
-        self.assertRaises(RuntimeError, cmp, y, x)
-        self.assertTrue(y is not x)
-        self.assertTrue(y[0] is not x[0])
-        self.assertTrue(y[0][0] is y)
+        for op in comparisons:
+            self.assertRaises(RuntimeError, op, y, x)
+        self.assertIsNot(y, x)
+        self.assertIsNot(y[0], x[0])
+        self.assertIs(y[0][0], y)
 
     def test_deepcopy_dict(self):
         x = {"foo": [1, 2], "bar": 3}
         y = copy.deepcopy(x)
         self.assertEqual(y, x)
-        self.assertTrue(x is not y)
-        self.assertTrue(x["foo"] is not y["foo"])
+        self.assertIsNot(x, y)
+        self.assertIsNot(x["foo"], y["foo"])
 
     def test_deepcopy_reflexive_dict(self):
         x = {}
         x['foo'] = x
         y = copy.deepcopy(x)
-        self.assertRaises(RuntimeError, cmp, y, x)
-        self.assertTrue(y is not x)
-        self.assertTrue(y['foo'] is y)
+        for op in order_comparisons:
+            self.assertRaises(TypeError, op, y, x)
+        for op in equality_comparisons:
+            self.assertRaises(RuntimeError, op, y, x)
+        self.assertIsNot(y, x)
+        self.assertIs(y['foo'], y)
         self.assertEqual(len(y), 1)
 
     def test_deepcopy_keepalive(self):
         memo = {}
-        x = 42
+        x = []
         y = copy.deepcopy(x, memo)
-        self.assertTrue(memo[id(x)] is x)
+        self.assertIs(memo[id(memo)][0], x)
+
+    def test_deepcopy_dont_memo_immutable(self):
+        memo = {}
+        x = [1, 2, 3, 4]
+        y = copy.deepcopy(x, memo)
+        self.assertEqual(y, x)
+        # There's the entry for the new list, and the keep alive.
+        self.assertEqual(len(memo), 2)
+
+        memo = {}
+        x = [(1, 2)]
+        y = copy.deepcopy(x, memo)
+        self.assertEqual(y, x)
+        # Tuples with immutable contents are immutable for deepcopy.
+        self.assertEqual(len(memo), 2)
 
     def test_deepcopy_inst_vanilla(self):
         class C:
             def __init__(self, foo):
                 self.foo = foo
-            def __cmp__(self, other):
-                return cmp(self.foo, other.foo)
+            def __eq__(self, other):
+                return self.foo == other.foo
         x = C([42])
         y = copy.deepcopy(x)
         self.assertEqual(y, x)
-        self.assertTrue(y.foo is not x.foo)
+        self.assertIsNot(y.foo, x.foo)
 
     def test_deepcopy_inst_deepcopy(self):
         class C:
@@ -332,13 +386,13 @@ class TestCopy(unittest.TestCase):
                 self.foo = foo
             def __deepcopy__(self, memo):
                 return C(copy.deepcopy(self.foo, memo))
-            def __cmp__(self, other):
-                return cmp(self.foo, other.foo)
+            def __eq__(self, other):
+                return self.foo == other.foo
         x = C([42])
         y = copy.deepcopy(x)
         self.assertEqual(y, x)
-        self.assertTrue(y is not x)
-        self.assertTrue(y.foo is not x.foo)
+        self.assertIsNot(y, x)
+        self.assertIsNot(y.foo, x.foo)
 
     def test_deepcopy_inst_getinitargs(self):
         class C:
@@ -346,13 +400,13 @@ class TestCopy(unittest.TestCase):
                 self.foo = foo
             def __getinitargs__(self):
                 return (self.foo,)
-            def __cmp__(self, other):
-                return cmp(self.foo, other.foo)
+            def __eq__(self, other):
+                return self.foo == other.foo
         x = C([42])
         y = copy.deepcopy(x)
         self.assertEqual(y, x)
-        self.assertTrue(y is not x)
-        self.assertTrue(y.foo is not x.foo)
+        self.assertIsNot(y, x)
+        self.assertIsNot(y.foo, x.foo)
 
     def test_deepcopy_inst_getstate(self):
         class C:
@@ -360,13 +414,13 @@ class TestCopy(unittest.TestCase):
                 self.foo = foo
             def __getstate__(self):
                 return {"foo": self.foo}
-            def __cmp__(self, other):
-                return cmp(self.foo, other.foo)
+            def __eq__(self, other):
+                return self.foo == other.foo
         x = C([42])
         y = copy.deepcopy(x)
         self.assertEqual(y, x)
-        self.assertTrue(y is not x)
-        self.assertTrue(y.foo is not x.foo)
+        self.assertIsNot(y, x)
+        self.assertIsNot(y.foo, x.foo)
 
     def test_deepcopy_inst_setstate(self):
         class C:
@@ -374,13 +428,13 @@ class TestCopy(unittest.TestCase):
                 self.foo = foo
             def __setstate__(self, state):
                 self.foo = state["foo"]
-            def __cmp__(self, other):
-                return cmp(self.foo, other.foo)
+            def __eq__(self, other):
+                return self.foo == other.foo
         x = C([42])
         y = copy.deepcopy(x)
         self.assertEqual(y, x)
-        self.assertTrue(y is not x)
-        self.assertTrue(y.foo is not x.foo)
+        self.assertIsNot(y, x)
+        self.assertIsNot(y.foo, x.foo)
 
     def test_deepcopy_inst_getstate_setstate(self):
         class C:
@@ -390,13 +444,19 @@ class TestCopy(unittest.TestCase):
                 return self.foo
             def __setstate__(self, state):
                 self.foo = state
-            def __cmp__(self, other):
-                return cmp(self.foo, other.foo)
+            def __eq__(self, other):
+                return self.foo == other.foo
         x = C([42])
         y = copy.deepcopy(x)
         self.assertEqual(y, x)
-        self.assertTrue(y is not x)
-        self.assertTrue(y.foo is not x.foo)
+        self.assertIsNot(y, x)
+        self.assertIsNot(y.foo, x.foo)
+        # State with boolean value is false (issue #25718)
+        x = C([])
+        y = copy.deepcopy(x)
+        self.assertEqual(y, x)
+        self.assertIsNot(y, x)
+        self.assertIsNot(y.foo, x.foo)
 
     def test_deepcopy_reflexive_inst(self):
         class C:
@@ -404,8 +464,8 @@ class TestCopy(unittest.TestCase):
         x = C()
         x.foo = x
         y = copy.deepcopy(x)
-        self.assertTrue(y is not x)
-        self.assertTrue(y.foo is y)
+        self.assertIsNot(y, x)
+        self.assertIs(y.foo, y)
 
     # _reconstruct()
 
@@ -415,9 +475,9 @@ class TestCopy(unittest.TestCase):
                 return ""
         x = C()
         y = copy.copy(x)
-        self.assertTrue(y is x)
+        self.assertIs(y, x)
         y = copy.deepcopy(x)
-        self.assertTrue(y is x)
+        self.assertIs(y, x)
 
     def test_reconstruct_nostate(self):
         class C(object):
@@ -426,24 +486,23 @@ class TestCopy(unittest.TestCase):
         x = C()
         x.foo = 42
         y = copy.copy(x)
-        self.assertTrue(y.__class__ is x.__class__)
+        self.assertIs(y.__class__, x.__class__)
         y = copy.deepcopy(x)
-        self.assertTrue(y.__class__ is x.__class__)
+        self.assertIs(y.__class__, x.__class__)
 
     def test_reconstruct_state(self):
         class C(object):
             def __reduce__(self):
                 return (C, (), self.__dict__)
-            def __cmp__(self, other):
-                return cmp(self.__dict__, other.__dict__)
-            __hash__ = None # Silence Py3k warning
+            def __eq__(self, other):
+                return self.__dict__ == other.__dict__
         x = C()
         x.foo = [42]
         y = copy.copy(x)
         self.assertEqual(y, x)
         y = copy.deepcopy(x)
         self.assertEqual(y, x)
-        self.assertTrue(y.foo is not x.foo)
+        self.assertIsNot(y.foo, x.foo)
 
     def test_reconstruct_state_setstate(self):
         class C(object):
@@ -451,16 +510,15 @@ class TestCopy(unittest.TestCase):
                 return (C, (), self.__dict__)
             def __setstate__(self, state):
                 self.__dict__.update(state)
-            def __cmp__(self, other):
-                return cmp(self.__dict__, other.__dict__)
-            __hash__ = None # Silence Py3k warning
+            def __eq__(self, other):
+                return self.__dict__ == other.__dict__
         x = C()
         x.foo = [42]
         y = copy.copy(x)
         self.assertEqual(y, x)
         y = copy.deepcopy(x)
         self.assertEqual(y, x)
-        self.assertTrue(y.foo is not x.foo)
+        self.assertIsNot(y.foo, x.foo)
 
     def test_reconstruct_reflexive(self):
         class C(object):
@@ -468,8 +526,8 @@ class TestCopy(unittest.TestCase):
         x = C()
         x.foo = x
         y = copy.deepcopy(x)
-        self.assertTrue(y is not x)
-        self.assertTrue(y.foo is y)
+        self.assertIsNot(y, x)
+        self.assertIs(y.foo, y)
 
     # Additions for Python 2.3 and pickle protocol 2
 
@@ -477,37 +535,35 @@ class TestCopy(unittest.TestCase):
         class C(list):
             def __reduce__(self):
                 return (C, (), self.__dict__, iter(self))
-            def __cmp__(self, other):
-                return (cmp(list(self), list(other)) or
-                        cmp(self.__dict__, other.__dict__))
-            __hash__ = None # Silence Py3k warning
+            def __eq__(self, other):
+                return (list(self) == list(other) and
+                        self.__dict__ == other.__dict__)
         x = C([[1, 2], 3])
         y = copy.copy(x)
         self.assertEqual(x, y)
-        self.assertTrue(x is not y)
-        self.assertTrue(x[0] is y[0])
+        self.assertIsNot(x, y)
+        self.assertIs(x[0], y[0])
         y = copy.deepcopy(x)
         self.assertEqual(x, y)
-        self.assertTrue(x is not y)
-        self.assertTrue(x[0] is not y[0])
+        self.assertIsNot(x, y)
+        self.assertIsNot(x[0], y[0])
 
     def test_reduce_5tuple(self):
         class C(dict):
             def __reduce__(self):
-                return (C, (), self.__dict__, None, self.iteritems())
-            def __cmp__(self, other):
-                return (cmp(dict(self), list(dict)) or
-                        cmp(self.__dict__, other.__dict__))
-            __hash__ = None # Silence Py3k warning
+                return (C, (), self.__dict__, None, self.items())
+            def __eq__(self, other):
+                return (dict(self) == dict(other) and
+                        self.__dict__ == other.__dict__)
         x = C([("foo", [1, 2]), ("bar", 3)])
         y = copy.copy(x)
         self.assertEqual(x, y)
-        self.assertTrue(x is not y)
-        self.assertTrue(x["foo"] is y["foo"])
+        self.assertIsNot(x, y)
+        self.assertIs(x["foo"], y["foo"])
         y = copy.deepcopy(x)
         self.assertEqual(x, y)
-        self.assertTrue(x is not y)
-        self.assertTrue(x["foo"] is not y["foo"])
+        self.assertIsNot(x, y)
+        self.assertIsNot(x["foo"], y["foo"])
 
     def test_copy_slots(self):
         class C(object):
@@ -515,7 +571,7 @@ class TestCopy(unittest.TestCase):
         x = C()
         x.foo = [42]
         y = copy.copy(x)
-        self.assertTrue(x.foo is y.foo)
+        self.assertIs(x.foo, y.foo)
 
     def test_deepcopy_slots(self):
         class C(object):
@@ -524,7 +580,7 @@ class TestCopy(unittest.TestCase):
         x.foo = [42]
         y = copy.deepcopy(x)
         self.assertEqual(x.foo, y.foo)
-        self.assertTrue(x.foo is not y.foo)
+        self.assertIsNot(x.foo, y.foo)
 
     def test_deepcopy_dict_subclass(self):
         class C(dict):
@@ -532,16 +588,16 @@ class TestCopy(unittest.TestCase):
                 if not d:
                     d = {}
                 self._keys = list(d.keys())
-                dict.__init__(self, d)
+                super().__init__(d)
             def __setitem__(self, key, item):
-                dict.__setitem__(self, key, item)
+                super().__setitem__(key, item)
                 if key not in self._keys:
                     self._keys.append(key)
         x = C(d={'foo':0})
         y = copy.deepcopy(x)
         self.assertEqual(x, y)
         self.assertEqual(x._keys, y._keys)
-        self.assertTrue(x is not y)
+        self.assertIsNot(x, y)
         x['bar'] = 1
         self.assertNotEqual(x, y)
         self.assertNotEqual(x._keys, y._keys)
@@ -554,8 +610,8 @@ class TestCopy(unittest.TestCase):
         y = copy.copy(x)
         self.assertEqual(list(x), list(y))
         self.assertEqual(x.foo, y.foo)
-        self.assertTrue(x[0] is y[0])
-        self.assertTrue(x.foo is y.foo)
+        self.assertIs(x[0], y[0])
+        self.assertIs(x.foo, y.foo)
 
     def test_deepcopy_list_subclass(self):
         class C(list):
@@ -565,8 +621,8 @@ class TestCopy(unittest.TestCase):
         y = copy.deepcopy(x)
         self.assertEqual(list(x), list(y))
         self.assertEqual(x.foo, y.foo)
-        self.assertTrue(x[0] is not y[0])
-        self.assertTrue(x.foo is not y.foo)
+        self.assertIsNot(x[0], y[0])
+        self.assertIsNot(x.foo, y.foo)
 
     def test_copy_tuple_subclass(self):
         class C(tuple):
@@ -583,13 +639,13 @@ class TestCopy(unittest.TestCase):
         self.assertEqual(tuple(x), ([1, 2], 3))
         y = copy.deepcopy(x)
         self.assertEqual(tuple(y), ([1, 2], 3))
-        self.assertTrue(x is not y)
-        self.assertTrue(x[0] is not y[0])
+        self.assertIsNot(x, y)
+        self.assertIsNot(x[0], y[0])
 
     def test_getstate_exc(self):
         class EvilState(object):
             def __getstate__(self):
-                raise ValueError, "ain't got no stickin' state"
+                raise ValueError("ain't got no stickin' state")
         self.assertRaises(ValueError, copy.copy, EvilState())
 
     def test_copy_function(self):
@@ -612,10 +668,10 @@ class TestCopy(unittest.TestCase):
         obj = C()
         x = weakref.ref(obj)
         y = _copy(x)
-        self.assertTrue(y is x)
+        self.assertIs(y, x)
         del obj
         y = _copy(x)
-        self.assertTrue(y is x)
+        self.assertIs(y, x)
 
     def test_copy_weakref(self):
         self._check_weakref(copy.copy)
@@ -626,12 +682,12 @@ class TestCopy(unittest.TestCase):
     def _check_copy_weakdict(self, _dicttype):
         class C(object):
             pass
-        a, b, c, d = [C() for i in xrange(4)]
+        a, b, c, d = [C() for i in range(4)]
         u = _dicttype()
         u[a] = b
         u[c] = d
         v = copy.copy(u)
-        self.assertFalse(v is u)
+        self.assertIsNot(v, u)
         self.assertEqual(v, u)
         self.assertEqual(v[a], b)
         self.assertEqual(v[c], d)
@@ -653,7 +709,7 @@ class TestCopy(unittest.TestCase):
         class C(object):
             def __init__(self, i):
                 self.i = i
-        a, b, c, d = [C(i) for i in xrange(4)]
+        a, b, c, d = [C(i) for i in range(4)]
         u = weakref.WeakKeyDictionary()
         u[a] = b
         u[c] = d
@@ -661,8 +717,8 @@ class TestCopy(unittest.TestCase):
         v = copy.deepcopy(u)
         self.assertNotEqual(v, u)
         self.assertEqual(len(v), 2)
-        self.assertFalse(v[a] is b)
-        self.assertFalse(v[c] is d)
+        self.assertIsNot(v[a], b)
+        self.assertIsNot(v[c], d)
         self.assertEqual(v[a].i, b.i)
         self.assertEqual(v[c].i, d.i)
         del c
@@ -672,7 +728,7 @@ class TestCopy(unittest.TestCase):
         class C(object):
             def __init__(self, i):
                 self.i = i
-        a, b, c, d = [C(i) for i in xrange(4)]
+        a, b, c, d = [C(i) for i in range(4)]
         u = weakref.WeakValueDictionary()
         u[a] = b
         u[c] = d
@@ -681,12 +737,12 @@ class TestCopy(unittest.TestCase):
         self.assertNotEqual(v, u)
         self.assertEqual(len(v), 2)
         (x, y), (z, t) = sorted(v.items(), key=lambda pair: pair[0].i)
-        self.assertFalse(x is a)
+        self.assertIsNot(x, a)
         self.assertEqual(x.i, a.i)
-        self.assertTrue(y is b)
-        self.assertFalse(z is c)
+        self.assertIs(y, b)
+        self.assertIsNot(z, c)
         self.assertEqual(z.i, c.i)
-        self.assertTrue(t is d)
+        self.assertIs(t, d)
         del x, y, z, t
         del d
         self.assertEqual(len(v), 1)
@@ -699,14 +755,14 @@ class TestCopy(unittest.TestCase):
         f.b = f.m
         g = copy.deepcopy(f)
         self.assertEqual(g.m, g.b)
-        self.assertTrue(g.b.im_self is g)
+        self.assertIs(g.b.__self__, g)
         g.b()
 
 
 def global_foo(x, y): return x+y
 
 def test_main():
-    test_support.run_unittest(TestCopy)
+    support.run_unittest(TestCopy)
 
 if __name__ == "__main__":
     test_main()
